import os
from typing import Any

from click import ClickException
import json
import jsonschema
import ray

from anyscale.util import slugify


# Pathnames specific to Ray's project directory structure.
PROJECT_ID_BASENAME = "project-id"
RAY_PROJECT_DIRECTORY = "ray-project"


def load_project_or_throw() -> Any:
    try:
        project = ray.projects.ProjectDefinition(os.getcwd())
    except (jsonschema.exceptions.ValidationError, ValueError) as e:
        raise ClickException(e)  # type: ignore

    dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(dir, "anyscale_schema.json")) as f:
        schema = json.load(f)

    # Validate the project file.
    try:
        jsonschema.validate(instance=project.config, schema=schema)
    except (jsonschema.exceptions.ValidationError, ValueError) as e:
        raise ClickException(e)  # type: ignore

    # Normalize project name
    project.config["name"] = slugify(project.config["name"])

    return project


def get_project_id(project_dir: str) -> int:
    """
    Args:
        project_dir: Project root directory.

    Returns:
        The ID of the associated Project in the database.

    Raises:
        ValueError: If the current project directory does
            not contain a project ID.
    """
    project_id_filename = os.path.join(
        project_dir, RAY_PROJECT_DIRECTORY, PROJECT_ID_BASENAME
    )
    if not os.path.isfile(project_id_filename):
        raise ClickException(
            "Ray project in {} not registered yet. "
            "Did you run 'any project create'?".format(project_dir)
        )
    with open(project_id_filename, "r") as f:
        project_id = f.read()
    try:
        result = int(project_id)
    except ValueError:
        raise ClickException(
            "{} does not contain a valid project ID".format(project_id_filename)
        )
    return result



import os
from typing import Optional

AWS_PROFILE = None

ANYSCALE_PRODUCTION_NAME = "anyscale.dev"

if "ANYSCALE_HOST" in os.environ:
    ANYSCALE_HOST = os.environ["ANYSCALE_HOST"]
else:
    # The production server.
    ANYSCALE_HOST = "https://" + ANYSCALE_PRODUCTION_NAME

# Global variable that contains the server session token.
CLI_TOKEN: Optional[str] = None

# Restic snapshot repo
TEST_MODE = False
SNAPSHOT_REPO = "s3:s3.amazonaws.com/anyscale-snapshots/internal"

SNAPSHOT_REPO_PASSWORD = "program_the_cloud"



import os

import anyscale.util


def report(message: str) -> None:
    """Show a message in the Anyscale Dashboard."""

    session_command_id = os.environ.get("ANYSCALE_SESSION_COMMAND_ID")

    if not session_command_id:
        raise RuntimeError(
            "Trying to use anyscale.report in a command " "not started by anyscale"
        )

    anyscale.util.send_json_request(
        "session_report_command",
        {"session_command_id": int(session_command_id), "message": message},
        post=True,
    )



import json
import logging
import requests
import os
import shutil
import subprocess
import time
import sys
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import click
from subprocess import STDOUT
import ray.ray_constants
import yaml

from anyscale.project import get_project_id
from anyscale.util import (
    confirm,
    send_json_request,
    get_cluster_config,
    get_requirements,
)
from anyscale.conf import SNAPSHOT_REPO, SNAPSHOT_REPO_PASSWORD


logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)


class SnapshotRunner:
    def __init__(self) -> None:
        self.snapshot_home = os.path.expanduser("~/.anyscale/snapshot")
        self.password = SNAPSHOT_REPO_PASSWORD
        self.snapshot_repo = SNAPSHOT_REPO

        # Get the right restic executable path depending on the OS.
        current_dir = os.path.dirname(os.path.realpath(__file__))
        if sys.platform.startswith("linux"):
            self.snapshot_executable = os.path.join(current_dir, "restic-linux")
        elif sys.platform.startswith("darwin"):
            self.snapshot_executable = os.path.join(current_dir, "restic-darwin")
        else:
            raise NotImplementedError(
                "Restic snapshot not supported on platform {}".format(sys.platform)
            )

    def setup(self) -> None:
        proc = subprocess.Popen(
            [self.snapshot_executable, "init", "-r", self.snapshot_repo],
            env={"RESTIC_PASSWORD": self.password},
        )

        proc.communicate()

    def upload(
        self,
        snapshot_uuid: str,
        project_dir: str,
        additional_files: List[str],
        files_only: bool,
        creds: Dict[str, str],
    ) -> None:
        """This function takes and upload a snapshot to S3. The snapshot
        consists of all the items in project_dir (if specified) and any
        additional_files specified.

        Args:
            snapshot_uuid: uuid of the snapshot to be uploaded to restic repo.
            project_dir: directory of the anyscale project.
            additional_files: additional files to be included with the snapshot (if any).
            creds: credentials used to access S3.
        """
        additional_files_list = tempfile.NamedTemporaryFile()
        with open(additional_files_list.name, "w") as temp:
            for f in additional_files:
                temp.write(f + "\n")

        new_env = os.environ.copy()
        new_env = {**new_env, **creds}
        new_env["RESTIC_PASSWORD"] = self.password
        print(json.dumps({"snapshot_uuid": snapshot_uuid}))
        proc = subprocess.Popen(
            [
                self.snapshot_executable,
                "-r",
                self.snapshot_repo,
                "backup",
                "--files-from",
                additional_files_list.name,
                "" if files_only else project_dir,
                "--tag",
                snapshot_uuid,
                "--tag",
                "project_dir=" + project_dir,
            ],
            env=new_env,
            stderr=STDOUT,
        )
        proc.communicate()

    def download(
        self, name: str, target_directory: Union[str, Path], creds: Dict[str, str],
    ) -> None:
        """This function downloads a snapshot from S3 to a target directory.
        The snapshot could be of a project, in which case it will have all
        project files and possibly output_files. Or it could be of any files
        on the remote cluster.
        Args:
            name: id of the snapshot to download.
            target_directory: directory to download the snapshot to.
            creds: credentials used to download the snapshot.
        """
        new_env = os.environ.copy()
        new_env = {**new_env, **creds}
        new_env["RESTIC_PASSWORD"] = self.password

        restic_id, project_dir, additional_files = self.get_restic_info(name, new_env)

        with tempfile.TemporaryDirectory() as snapshot_dir:
            proc = subprocess.Popen(
                [
                    self.snapshot_executable,
                    "-r",
                    self.snapshot_repo,
                    "restore",
                    "--tag",
                    name,
                    "latest",
                    "--target",
                    snapshot_dir,
                ],
                env=new_env,
                stderr=STDOUT,
            )

            proc.communicate()

            # Convert absolute path used by restic to relative path
            # Strip the leading "/"
            relative_directory = project_dir[1:]
            current_dir = os.path.join(snapshot_dir, relative_directory)
            if os.path.exists(current_dir):
                # Copy the src directory to dst directory recursively
                # and overwrite if the file already exists
                proc = subprocess.Popen(["rsync", "-a", current_dir, target_directory])
                proc.communicate()

            # Move additional files
            for additional_file in additional_files:
                proc = subprocess.Popen(
                    [
                        "rsync",
                        "-a",
                        os.path.join(snapshot_dir, additional_file[1:]),
                        target_directory,
                    ]
                )
                proc.communicate()

    def delete(self, name: str, creds: Dict[str, str]) -> None:
        new_env = os.environ.copy()
        new_env = {**new_env, **creds}
        new_env["RESTIC_PASSWORD"] = self.password

        # Getting restic snapshot ID from snapshot_uuid
        # Need this because we cannot delete snapshot with --tag
        restic_id, _, _ = self.get_restic_info(name, new_env)

        # Need this because if a deletion fails, restic repo will be locked.
        proc = subprocess.Popen(
            [self.snapshot_executable, "-r", self.snapshot_repo, "unlock"],
            env=new_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Actual deletion
        proc = subprocess.Popen(
            [
                self.snapshot_executable,
                "-r",
                self.snapshot_repo,
                "forget",
                restic_id,
                "--prune",
            ],
            env=new_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output, error = proc.communicate()
        if proc.returncode != 0:
            print(output, error)

    def describe(self, name: str, creds: Dict[str, str]) -> List[Dict[str, str]]:
        new_env = os.environ.copy()
        new_env = {**new_env, **creds}
        new_env["RESTIC_PASSWORD"] = self.password

        restic_id, project_dir, _ = self.get_restic_info(name, new_env)
        proc = subprocess.Popen(
            [
                self.snapshot_executable,
                "-r",
                self.snapshot_repo,
                "ls",
                restic_id,
                "--recursive",
                project_dir,
                "--json",
            ],
            env=new_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output, error = proc.communicate()
        if proc.returncode != 0:
            print(output, error)

        snapshot_items = []
        for element in output.strip().split(b"\n"):
            snapshot_item = json.loads(element)
            path = snapshot_item.get("path")
            if path == project_dir:
                continue
            if path:
                snapshot_item["path"] = path[len(project_dir) :]
                snapshot_items.append(snapshot_item)

        return snapshot_items

    def get_restic_info(
        self, tag: str, new_env: Dict[str, str]
    ) -> Tuple[str, str, List[str]]:
        # Getting restic snapshot ID from snapshot_uuid
        # Need this because we cannot delete snapshot with --tag
        proc = subprocess.Popen(
            [
                self.snapshot_executable,
                "-r",
                self.snapshot_repo,
                "snapshots",
                "--tag",
                tag,
                "--json",
            ],
            env=new_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output, error = proc.communicate()
        if proc.returncode != 0:
            print(output, error)

        return parse_restic_snapshots_output(output)


snapshot_runner = SnapshotRunner()


def parse_restic_snapshots_output(output: bytes) -> Tuple[str, str, List[str]]:
    """
    This function takes an ouput from the "restic snapshots" cmd
    and extract restic snapshot id and its project_dir
    """
    snapshots_info = json.loads(output.decode())
    assert snapshots_info and len(snapshots_info) >= 1, "Snapshot not found"
    snapshot_info = snapshots_info[0]
    project_dir_prefix = "project_dir="
    project_dir = ""
    for tag in snapshot_info["tags"]:
        if tag.startswith(project_dir_prefix):
            project_dir = tag[len(project_dir_prefix) :]
    assert (
        len(project_dir) > 0
    ), "Project directory not found. Maybe it's not tagged to the snapshot?"

    output_files = []
    for path in snapshot_info["paths"]:
        if not path.startswith(os.path.abspath(project_dir)):
            output_files.append(path)

    return snapshot_info["short_id"], project_dir, output_files


def copy_file(to_s3: bool, source: str, target: Any, download: bool) -> None:
    """Copy a file.

    The file source or target may be on S3.

    Args:
        to_s3 (bool): If this is True, then copy to/from S3, else the local
            disk. If this is True, then the file source or target will be a
            presigned URL to which GET or POST HTTP requests can be sent.
        source (str or S3 URL): Source file local pathname or S3 GET URL. If
            this is an S3 URL, target is assumed to be a local pathname.
        target (str or S3 URL): Target file local pathname or S3 URL with POST
            credentials. If this is an S3 URL, source is assumed to be a local
            pathname.
        download (bool): If this is True, then this will upload from source to
            target, else this will download.
    """
    try:
        if to_s3:
            if download:
                with open(target, "wb") as f:
                    response = requests.get(source)
                    for block in response.iter_content(1024):
                        f.write(block)
            else:
                with open(source, "rb") as f:
                    files = {"file": ("object", f)}
                    resp = requests.post(
                        target["url"], data=target["fields"], files=files
                    )
                    assert resp.ok, resp.text
        else:
            shutil.copyfile(source, target)
    except (OSError, AssertionError) as e:
        logger.warn("Failed to copy file %s , aborting", source)
        raise e


def create_snapshot(
    project_definition: Any,
    yes: bool,
    description: Optional[str] = None,
    include_output_files: bool = False,
    additional_files: List[str] = [],
    files_only: bool = False,
    tags: List[str] = [],
) -> str:
    """Create a snapshot of a project.

    Args:
        project_definition: Project definition.
        yes: Don't ask for confirmation.
        description: An optional description of the snapshot.
        include_output_files: Whether or not to include output files.
        additional_files: Additional files to include in the snapshot.
        files_only: Only include additional files in the snapshot.
        tags: Tags for the snapshot.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
        Exception: If saving the snapshot files fails.
    """
    # Find and validate the current project ID.
    project_dir = project_definition.root
    project_id = get_project_id(project_dir)

    if include_output_files:
        for output_file in project_definition.config.get("output_files", []):
            if output_file[0] == "/" and not output_file.startswith(project_dir):
                additional_files.append(output_file)

    cluster_config = get_cluster_config(
        os.path.join(project_dir, project_definition.cluster_yaml())
    )

    if "requirements" in project_definition.config.get("environment", {}):
        requirements = get_requirements(
            os.path.join(
                project_dir, project_definition.config["environment"]["requirements"]
            )
        )
    else:
        requirements = ""

    resp = send_json_request(
        "snapshot_create",
        {
            "project_id": project_id,
            "project_config": json.dumps(project_definition.config),
            "cluster_config": json.dumps(cluster_config),
            "requirements": requirements,
            "description": description if description else "",
            "tags": tags,
        },
        post=True,
    )

    snapshot_uuid: str = resp["uuid"]
    creds = resp["credentials"]

    try:
        snapshot_runner.upload(
            snapshot_uuid, project_dir, additional_files, files_only, creds
        )
    except (OSError, AssertionError) as e:
        # Tell the server to delete the snapshot.
        delete_snapshot(project_dir, snapshot_uuid, True)
        raise e
    return snapshot_uuid


def delete_snapshot(project_dir: str, uuid: str, yes: bool) -> None:
    """Delete the snapshot(s) with the given name.

    Delete the snapshot data from disk and the metadata from the database.

    Args:
        project_dir: Project root directory.
        uuid: The UUID of the snapshot to delete.
        yes: Don't ask for confirmation.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)

    # Get the snapshots with the requested name.
    resp = send_json_request(
        "snapshot_list", {"project_id": project_id, "snapshot_uuid": uuid}
    )
    snapshots = resp["snapshots"]
    if snapshots:
        snapshots_str = "\n".join(snapshot["uuid"] for snapshot in snapshots)
        confirm("Delete snapshots?\n{}".format(snapshots_str), yes)
        snapshot_uuids = [snapshot["uuid"] for snapshot in snapshots]
        resp = send_json_request(
            "snapshot_delete", {"snapshot_uuids": snapshot_uuids}, post=True
        )

        assert "credentials" in resp

        snapshot_runner = SnapshotRunner()
        for name in snapshot_uuids:
            snapshot_runner.delete(name, resp["credentials"])

    else:
        logger.warn("No snapshots found with UUID %s", uuid)


def describe_snapshot(uuid: str) -> Any:
    resp = send_json_request("snapshot_describe", {"snapshot_uuid": uuid})
    return resp


def list_snapshots(project_dir: str) -> List[str]:
    """List all snapshots associated with the given project.

    Args:
        project_dir: Project root directory.

    Returns:
        List of Snapshots for the current project.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)
    resp = send_json_request("snapshot_list", {"project_id": project_id})
    snapshots = resp["snapshots"]
    return [snapshot["uuid"] for snapshot in snapshots]


def get_snapshot_uuid(project_dir: str, snapshot_uuid: str) -> str:
    """Get a snapshot of the given project with the given name.

    Args:
        project_id: The ID of the project.
        snapshot_name: The name of the snapshot to get. If there are multiple
            snapshots with the same name, then the user will be prompted to
            choose a snapshot.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)
    resp = send_json_request(
        "snapshot_list", {"project_id": project_id, "snapshot_uuid": snapshot_uuid}
    )
    snapshots = resp["snapshots"]
    if len(snapshots) == 0:
        raise ValueError("No snapshots found with name {}".format(snapshot_uuid))
    snapshot_idx = 0
    if len(snapshots) > 1:
        print(
            "More than one snapshot found with UUID {}. "
            "Which do you want to use?".format(snapshot_uuid)
        )
        for i, snapshot in enumerate(snapshots):
            print("{}. {}".format(i + 1, snapshot["uuid"]))
        snapshot_idx = click.prompt(
            "Please enter a snapshot number from 1 to {}".format(len(snapshots)),
            type=int,
        )
        snapshot_idx -= 1
        if snapshot_idx < 0 or snapshot_idx > len(snapshots):
            raise ValueError("Snapshot index {} is out of range".format(snapshot_idx))
    result: str = snapshots[snapshot_idx]["uuid"]
    return result


def download_snapshot(
    name: str,
    credentials: Dict[str, str],
    target_directory: Optional[str] = None,
    overwrite: bool = False,
) -> str:
    """Download a snapshot to a local target directory.

    This will recreate the original directory structure of the snapshot. Only
    input files (those found in the project directory during the snapshot
    creation) will be downloaded, not output files specified in the
    project.yaml.

    Args:
        snapshot_info: Information about the snapshot as returned by
            snapshot_get.
        credentials: AWS credentials for accessing S3.
        target_directory: Directory this snapshot gets downloaded to.
            If None, the snapshot will be downloaded to a temporary directory.

    Returns:
        str: The directory where the snapshot has been downloaded to.
    """
    if not target_directory:
        target_directory = os.path.join(os.getcwd(), name)
    target_directory = os.path.abspath(target_directory)
    if os.path.exists(target_directory) and not overwrite:
        print(
            'Target directory already exists. Run with "--overwrite" flag to overwrite the target directory.'
        )
        return ""
    snapshot_runner.download(name, target_directory, credentials)

    return target_directory



import copy
import os
import subprocess
import sys
import xml.etree.ElementTree


class AutosyncRunner:
    def __init__(self) -> None:
        self.syncthing_home = os.path.expanduser("~/.anyscale/syncthing")
        self.syncthing_config_path = os.path.join(self.syncthing_home, "config.xml")
        # Make it so syncthing does not try to automatically update itself.
        # This also avoids configurations getting out of sync.
        # We should only roll out new versions deliberately after testing.
        self.env = os.environ.copy()
        self.env["STNOUPGRADE"] = "1"

        # Get the right syncthing executable path depending on the OS.
        current_dir = os.path.dirname(os.path.realpath(__file__))
        if sys.platform.startswith("linux"):
            self.syncthing_executable = os.path.join(current_dir, "syncthing-linux")
        elif sys.platform.startswith("darwin"):
            self.syncthing_executable = os.path.join(current_dir, "syncthing-macos")
        else:
            raise NotImplementedError(
                "Autosync not supported on platform {}".format(sys.platform)
            )

        # This line creates the configuration file for the autosync.
        assert (
            subprocess.check_call(
                [
                    self.syncthing_executable,
                    "-allow-newer-config",
                    "-generate",
                    self.syncthing_home,
                ],
                env=self.env,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            == 0
        )

        # Deactivate the user reporting.
        config = xml.etree.ElementTree.parse(self.syncthing_config_path)
        root = config.getroot()
        for child in root.getchildren():
            if child.tag == "options":
                for element in child.getchildren():
                    if element.tag == "urAccepted":
                        element.text = "-1"
        config.write(self.syncthing_config_path)

    def add_device(self, project_name: str, device_id: str) -> None:
        """Add a device to syncthing config to device and folder section.

        The syncthing configuration path used is in the SYNCTHING_CONFIG_PATH
        variable. If this device_id has already been added it will not be
        added again.

        Args:
            project_name: Name of the project we add the device to.
            device_id: ID of the device being added.
        """
        config = xml.etree.ElementTree.parse(self.syncthing_config_path)
        root = config.getroot()
        project_folder_element = None
        device_element = None
        for child in root.getchildren():
            if child.get("id") == project_name:
                project_folder_element = child
            if child.get("id") == device_id:
                device_element = child
        assert project_folder_element is not None
        # Add the device element to the top level configuration if it
        # is not already there.
        if not device_element:
            device_element = xml.etree.ElementTree.SubElement(
                root,
                "device",
                {
                    "compression": "metadata",
                    "id": device_id,
                    "introducedBy": "",
                    "introducer": "false",
                    "name": project_name,
                    "skipIntroductionRemovals": "false",
                },
            )
            properties = {
                "address": "dynamic",
                "paused": "false",
                "autoAcceptFolders": "false",
                "maxSendKbps": "0",
                "maxRecvKbps": "0",
                "maxRequestKiB": "0",
            }
            for key, value in properties.items():
                element = xml.etree.ElementTree.SubElement(device_element, key)
                element.text = value
            root.append(device_element)

        # We want to add the device element to the folder if it is
        # not already there. First check if it is, then add it if needed.
        folder_device_element_present = False
        for child in project_folder_element.getchildren():
            if project_folder_element.get("id") == device_id:
                folder_device_element_present = True
        if not folder_device_element_present:
            folder_device_element = xml.etree.ElementTree.SubElement(
                project_folder_element, "device", {"id": device_id}
            )
            project_folder_element.append(folder_device_element)

        config.write(self.syncthing_config_path)

    def add_or_update_project_folder(
        self, project_name: str, project_folder_path: str
    ) -> None:
        """Add a project folder to the syncthing config.xml.

        The syncthing configuration path used is in the SYNCTHING_CONFIG_PATH
        variable. This function will not add any devices to the folder.

        Args:
            project_name: Name of the project we add the folder
                for. This is used as an identifier for the folder.
            project_folder_path: Full path to the folder we want
                to add to syncthing.
        """
        config = xml.etree.ElementTree.parse(self.syncthing_config_path)
        root = config.getroot()
        default_folder_element = None
        project_folder_element = None
        for child in root.getchildren():
            if child.get("id") == "default":
                default_folder_element = child
            if child.get("id") == project_name:
                project_folder_element = child
        assert default_folder_element is not None
        if not project_folder_element:
            project_folder_element = copy.deepcopy(default_folder_element)
            project_folder_element.set("id", project_name)
            project_folder_element.set(
                "label", "Project Folder for {}".format(project_name)
            )
            project_folder_element.set("path", project_folder_path)
            project_folder_element.set("fsWatcherDelayS", "1")
            root.append(project_folder_element)
        else:
            project_folder_element.set("path", project_folder_path)
        config.write(self.syncthing_config_path)

    def get_device_id(self) -> str:
        return (
            subprocess.check_output(
                [
                    self.syncthing_executable,
                    "-allow-newer-config",
                    "-home",
                    self.syncthing_home,
                    "-device-id",
                ],
                env=self.env,
            )
            .decode()
            .strip()
        )

    def start_autosync(self, verbose: bool) -> None:
        command = [
            self.syncthing_executable,
            "-allow-newer-config",
            "-home",
            self.syncthing_home,
            "-no-browser",
        ]
        if not verbose:
            syncthing = subprocess.Popen(
                command,
                env=self.env,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
            )
            success = False
            while True:
                if syncthing.stdout:
                    line = syncthing.stdout.readline()
                    if not line:
                        break
                    if b"GUI and API listening on" in line:
                        address = line.decode().split()[-1]
                        print(
                            "Autosync is active while this program is running, "
                            "status is available at {}...".format(address)
                        )
                        success = True
                else:
                    print("Autosync is active while this program is running...")
            if not success:
                print(
                    "Autosync failed. Please retry with the --verbose flag to see the error."
                )
        else:
            syncthing = subprocess.Popen(command, env=self.env)
            syncthing.communicate()

    def kill_autosync(self) -> None:
        subprocess.check_output(["pkill syncthing || true"], shell=True)



ENDPOINTS = {
    "project_create": "/api/project/create",
    "project_create_from_github": "/api/v1/project/create_from_github",
    "project_list": "/api/project/list",
    "project_sessions": "/api/project/sessions",
    "session_autosync_add_device": "/api/session/autosync_add_device",
    "session_autosync_started": "/api/session/autosync_started",
    "session_commands": "/api/session/commands",
    "session_create": "/api/session/create",
    "session_describe": "/api/session/describe",
    "session_execute": "/api/v1/session/execute",
    "session_execution_logs": "/api/session/execution_logs",
    "session_monitor_logs": "/api/session/monitor_logs",
    "session_finish_command": "/api/session/finish_command",
    "session_kill_command": "/api/session/kill_command",
    "session_list": "/api/session/list",
    "session_overview": "/api/session/overview",
    "session_report_command": "/api/session/report_command",
    "session_startup_log": "/api/v1/session/startup_log",
    "session_snapshot": "/api/v1/session/snapshot",
    "session_start": "/api/session/start",
    "session_stop": "/api/session/stop",
    "session_sync": "/api/session/sync",
    "session_upload_command_logs": "/api/session/upload_command_logs",
    "snapshot_archive": "/api/snapshot/archive",
    "snapshot_create": "/api/snapshot/create",
    "snapshot_delete": "/api/snapshot/delete",
    "snapshot_describe": "/api/v1/snapshot/describe",
    "snapshot_get": "/api/snapshot/get",
    "snapshot_list": "/api/snapshot/list",
    "user_info": "/api/user/info",
    "user_login": "/api/user/login",
    "user_logout": "/api/user/logout",
    "user_register": "/api/user/register",
    "user_server_session_token": "/api/user/server_session_token",
    "user_get_temporary_aws_credentials": "/api/user/temporary_credentials",
    "user_aws_get_anyscale_account": "/api/v1/user/aws/anyscale_account",
    "user_setup_aws_cross_account_role": "/api/v1/user/aws/cross_account_role",
}
